/*********************************************************************
*
* ANSI C Example program:
*    contGeneration.c
*
* Example Category:
*    AO
*
* Description:
*    This example demonstrates how to output a continuous periodic
*    waveform using an internal sample clock.
*
* Instructions for Running:
*    1. Select the Physical Channel to correspond to where your signal
*       is output on the DAQ device.
*    2. Enter the Minimum and Maximum Voltage Ranges.
*    3. Enter the desired rate for the generation. The onboard sample
*       clock will operate at this rate.
*    4. Select the desired waveform type.
*    5. The rest of the parameters in the Waveform Information section
*       will affect the way the waveform is created, before it's sent
*       to the analog output of the board. Select the number of samples
*       per cycle and the total number of cycles to be used as waveform
*       data.
*
* Steps:
*    1. Create a task.
*    2. Create an Analog Output Voltage channel.
*    3. Define the update Rate for the Voltage generation. Additionally,
*       define the sample mode to be continuous.
*    4. Write the waveform to the output buffer.
*    5. Call the Start function.
*    6. Loop continuously for 10 seconds. Check for errors every
*       100 ms using the IsTaskDone function.
*    7. Call the Clear Task function to clear the Task.
*    8. Display an error if any.
*
* I/O Connections Overview:
*    Make sure your signal output terminal matches the Physical Channel
*    I/O Control. For further connection information, refer to your
*    hardware reference manual.
*
* Recommended use:
*    Call Configure, Write and Start functions.
*    Call IsDone function in a loop.
*    Call Stop function at the end.
*
*********************************************************************/
#include <NIDAQmx.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <unistd.h>

#define DAQmxErrChk(functionCall) { if( DAQmxFailed(error=(functionCall)) ) { goto Error; } }
#define PI	3.1415926535
#define BUFFERSIZE	4096
#define NUMCHANNELS	4
#define NUMAOCHANNELS	3
#define SMOOTH_FRAC	0.05

int TriangleWave(int number,float amp,float phase_offset,float64 *data);
int Trigger(int number, int width, uInt8 *data);
int BoxcarSmooth(int number, float64 *data, int width, float64 *data_sm);

int main(int argc, char *argv[])
{
    // Task parameters
    int32       error = 0;
    TaskHandle  mod_taskHandle = 0;
    int32       i,c; 
    char        errBuff[2048]={'\0'};

    // Channel parameters
    char        chan[] = "Dev1/ao0:3";
    float64     min = -10.0;
    float64     max = 10.0;

    // Timing parameters
    float64     sample_rate;

    // Data write parameters
    int32	data_size = BUFFERSIZE * NUMCHANNELS;
    float64     waveform[BUFFERSIZE], waveform2[BUFFERSIZE];
    uInt8	ttl[BUFFERSIZE];
    float64	data[data_size];
    int32       pointsWritten;
    float64     timeout = 10.0;
    //    FILE	*fp;
    float	phase_offset;

    // Clock parameters
    TaskHandle  clock_taskHandle = 0;
    float64 clock_freq;

    // Trigger parameters
    float	pulseWidth;
    float64	trigger_delay;
    TaskHandle  trigger_taskHandle = 0;
    float64 	dutyCycle;
    int32	lowTicks, highTicks;

    // Misc
    float       freq, amp;
    int32	samples_per_chan;
    int32	smooth_width;



	/* Reset the device */
	printf("Resetting device... ");
	DAQmxErrChk(DAQmxResetDevice("Dev1"));
	DAQmxErrChk(DAQmxResetDevice("Dev1"));
	DAQmxErrChk(DAQmxResetDevice("Dev1"));
	DAQmxErrChk(DAQmxResetDevice("Dev1"));
	printf("Done\n");


        /* read the command line to get the frequency */
        if (argc<2)
        {
                freq=1.;
        }
        else
        {
                sscanf(*++argv,"%f", &freq);
        }
	printf("Dither frequency: %0.1f Hz\n", freq);

        /* read the command line to get the amplide */
        if (argc<3)
        {
                amp=3.*0.1;	//in V
        }
        else
        {
                sscanf(*++argv,"%f", &amp);
        }
	printf("Amplitude: %0.3f V\n", amp);

        /* read the command line to get the pulse width */
        if (argc<4)
        {
                pulseWidth=0.5;	//in ms
        }
        else
        {
                sscanf(*++argv,"%f", &pulseWidth);
        }
	printf("Pulse width: %0.1f ms\n", pulseWidth);


        /* init the vectors */
        for (i=0; i<BUFFERSIZE; i++){
		waveform[i]=0;
		waveform2[i]=0;
	}
        for (i=0; i<data_size; i++){
		data[i]=0;
	}
	printf("Data size: %ld\n", data_size);


	/* set the number of samples per channel */
        samples_per_chan=BUFFERSIZE;
        printf("Number of samples per channel: %ld\n", samples_per_chan);

        /* make a triangle wave */
	phase_offset = -2*PI*SMOOTH_FRAC;
        TriangleWave(samples_per_chan,amp,phase_offset,waveform);

        /* smooth it by SMOOTH_FRAC */
        smooth_width=(int)(samples_per_chan*SMOOTH_FRAC);
        printf("Smoothing boxcar width per cycle: %ld (%.0f%%)\n", 
		smooth_width, SMOOTH_FRAC*100.);
        BoxcarSmooth(samples_per_chan,waveform,smooth_width,waveform2);

        /* what sample rate (in samples/second) do we need */
	/* to get the freq we want? */
	sample_rate=freq*samples_per_chan;
        printf("Sample rate: %f samples/ms\n", sample_rate/1000);

	/* the TTL pulse */
	Trigger(samples_per_chan,
		sample_rate/1000,
		ttl);

        /* write the data (grouped by channel) for multiple channels */
        for (i=0;i<samples_per_chan;i++){
	   for (c=0;c<NUMAOCHANNELS;c++)
              data[i+c*samples_per_chan]=waveform2[i];
           data[i+3*samples_per_chan]=ttl[i];
	}

	/* write the data to a file for posterity */
//	fp = fopen("waveform.dat", "w");
//        for (i=0; i<data_size; i++){
//		fprintf(fp,"%ld %f\n",i,data[i]);
//	}
//	fclose(fp);



	/*
	 * MODULATION signal
	 */


	/* Create the analog out task */
	DAQmxErrChk (DAQmxCreateTask("modulation",&mod_taskHandle));


	/* Create the channel(s) and add them to the task */
	DAQmxErrChk (
		DAQmxCreateAOVoltageChan(
			mod_taskHandle,
			chan,
			"",
			min,
			max,
			DAQmx_Val_Volts,
			NULL
		)
	);

	/* Set the source and rate of the clock and the number of samples */
	printf("Set the clock source and rate... ");
	DAQmxErrChk (
		DAQmxCfgSampClkTiming(
			mod_taskHandle,
			"Ctr0InternalOutput",
			sample_rate,
			DAQmx_Val_Rising,
			DAQmx_Val_ContSamps,
			samples_per_chan
		)
	);
	printf("Done\n");

	/* Write the data to the buffer */
	printf("Writing waveform data to the buffer... ");
	DAQmxErrChk (
		DAQmxWriteAnalogF64(
			mod_taskHandle,
			samples_per_chan,
			0,			// auto-start set to false
			timeout,
			DAQmx_Val_GroupByChannel,
			data,
			&pointsWritten,
			NULL
		)
	);
	printf("Done\n");
    	printf("Number of points written per channel:  %ld\n", pointsWritten);

	/* sync the the mod signal with the trigger */
	DAQmxErrChk (
	    DAQmxCfgDigEdgeStartTrig(
	       mod_taskHandle, 
	       "/Dev1/Ctr0InternalOutput", 
	       DAQmx_Val_Rising
	    )
	 );

	/* start the task */
	DAQmxErrChk (DAQmxStartTask(mod_taskHandle));


	/*
	 * CLOCK
	 */

	clock_freq = freq*samples_per_chan;
	printf("Clock frequency: %f Hz\n", clock_freq);

	/* create the clock task */
	DAQmxErrChk (DAQmxCreateTask("clock",&clock_taskHandle));

	/* setup the clock */
	DAQmxErrChk (
		DAQmxCreateCOPulseChanFreq(
			clock_taskHandle,
			"Dev1/ctr0",
			"",
			DAQmx_Val_Hz,
			DAQmx_Val_Low,
			0.0,
			clock_freq,
			0.5
		)
	);

	/* implicit timing */
	DAQmxErrChk (
	   DAQmxCfgImplicitTiming(
	      clock_taskHandle,
	      DAQmx_Val_ContSamps,
	      0
	   )
	);



	/*
	 * TRIGGER
	 */ 
	if ((dutyCycle = (pulseWidth/1000.) * clock_freq) < 1) {
	   dutyCycle=2;
	   printf("WARNING:  dutyCycle too low, setting to %f (%f ms)\n", 
	           dutyCycle, dutyCycle*1000/clock_freq);
	} else {
	   printf("dutyCycle: %f\n", dutyCycle);
	}

	trigger_delay = samples_per_chan/4.-smooth_width/2;

	highTicks=dutyCycle;
	lowTicks=samples_per_chan/2-highTicks;

	/* create the task */
	DAQmxErrChk (DAQmxCreateTask("trigger",&trigger_taskHandle));

	/* setup the trigger in terms of clock ticks */
	DAQmxErrChk (
	   DAQmxCreateCOPulseChanTicks(
	       trigger_taskHandle, 
	       "Dev1/ctr1", 
	       "", 
	       "Ctr0InternalOutput", 
	       DAQmx_Val_Low, 
	       trigger_delay, 
	       lowTicks,
	       highTicks
	    )
	 );

	DAQmxErrChk (
	   DAQmxCfgImplicitTiming(
	      trigger_taskHandle,
	      DAQmx_Val_ContSamps,
	      0
	   )
	);

	/* start the trigger when the clock starts */
	DAQmxErrChk (
	   DAQmxCfgDigEdgeStartTrig (
	      trigger_taskHandle, 
	      "Ctr0InternalOutput",
	      DAQmx_Val_Rising
	   )
	);

	DAQmxErrChk (DAQmxStartTask(trigger_taskHandle));


	/* start the clock, which starts the mod and trig */
	DAQmxErrChk (DAQmxStartTask(clock_taskHandle));


	printf("Press <ENTER> when ready to continue\n");
	getchar();

	DAQmxErrChk (DAQmxStopTask(mod_taskHandle));
	printf("Setting all channels to zero...");
        for (i=0; i<data_size; i++) data[i]=0;

	DAQmxErrChk (
		DAQmxWriteAnalogF64(
			mod_taskHandle,
			samples_per_chan,
			0,
			timeout,
			DAQmx_Val_GroupByChannel,
			data,
			&pointsWritten,
			NULL
		)
	);
	printf("Done\n");
	DAQmxErrChk (DAQmxStartTask(mod_taskHandle));

	sleep(1);

Error:
	/* if there was an error, get a more user-friendly message */
	if( DAQmxFailed(error) ){
	    printf("\n\nERROR ERROR ERROR\n\n");
	    DAQmxGetExtendedErrorInfo(errBuff,2048);
	}

	/* if the task has let go, stop and clear it */
	if( (mod_taskHandle!=0) && (trigger_taskHandle!=0) ) {
		printf("Stopping and Clearing Task(s)... ");
		DAQmxStopTask(clock_taskHandle);
		DAQmxStopTask(mod_taskHandle);
		DAQmxStopTask(trigger_taskHandle);
		DAQmxClearTask(clock_taskHandle);
		DAQmxClearTask(mod_taskHandle);
		DAQmxClearTask(trigger_taskHandle);
		printf("Done\n");
	}

	/* report the error if there was one */
	if( DAQmxFailed(error) )
	    printf("DAQmx Error: %s\n",errBuff);

	return 0;
}


/* function to create a triangle waveform */
int TriangleWave(int number,float amp, float phase_offset, float64 *data)
{
        int i;
        float64 phase;

        for (i=0;i<number;i++)
        {
                phase=2*PI*(float)i/(float)number + phase_offset;
		// data amp is centered on 5 (middle of DAQ range)
                *data++ = amp*(asin(sin(phase))/PI)+5.;
        }

        return 0;
}

/* function to create a trigger waveform */
int Trigger(int number,int width, uInt8 *data)
{
        int i;

        for (i=0;i<number;i++)
        {
		if ( (i>0) && (i<width)){
			*data++ = 1;
		} else {
			*data++ = 0;
		}
        }

        return 0;
}

/* boxcar smoothes any waveform */
int BoxcarSmooth(int number, float64 *data, int width, float64 *data_sm)
{
        float64 sum;
        int i,j,index;

        for (i=0;i<number;i++)
        {
                sum=0.0;
                for (j=0;j<width-1;j++)
                {
                        index=i+j-width/2;
                        if (index >= number) index-=number;
                        if (index < 0) index+=number;
                        sum+=data[index];
                }
                sum/=width;
                *data_sm++=sum;
        }

        return 0;
}

