/************************************************************************/
/* champ_chara_server.c							*/
/*                                                                      */
/* Contains the code required to make CHAMP a chara server on the 	*/
/* system.								*/
/************************************************************************/
/*                                                                      */
/*                    CHARA ARRAY SERVER LIB   				*/
/*                 Based on the CHARA User Interface			*/
/*                 Based on the SUSI User Interface			*/
/*		In turn based on the CHIP User interface		*/
/*                                                                      */
/*            Center for High Angular Resolution Astronomy              */
/*              Mount Wilson Observatory, CA 91001, USA			*/
/*                                                                      */
/* Telephone: 1-626-796-5405                                            */
/* Fax      : 1-626-796-6717                                            */
/* email    : theo@chara.gsu.edu                                        */
/* WWW      : http://www.chara.gsu.edu			                */
/*                                                                      */
/* (C) This source code and its associated executable                   */
/* program(s) are copyright.                                            */
/*                                                                      */
/************************************************************************/
/*                                                                      */
/* Author : Theo ten Brummelaar 		                        */
/* Date   : June 2011							*/
/************************************************************************/

/* This definition is required for servers and is used in the cliserv.h file */

#define SERVER

/* Now we include the header as a server. */

#undef __CLISERV__
#include <cliserv.h>

/* The chara message definitions */

#include "chara_messages.h"
#include "fitsio.h"
#include "chara.h"

/* A global for information used by example message function */

static int reference_cart = -1;

STAR_DATA star;

/* Local declarations */

int message_astromod_set_use_astromod_lib(struct smessage *message);
int message_observe_ople_number(struct smessage *message);
int message_astromod_init(struct smessage *message);
int message_astromod_set_reference_cart(struct smessage *message);
int message_astromod_which_reference_cart(struct smessage *message);
int message_astromod_set_active_scopes(struct smessage *message);
int message_astromod_set_star(struct smessage *message);
int message_observe_star_acquired(struct smessage *message);
int message_astromod_set_obs_wavelength(struct smessage *message);
void stop_server(int sig);
int client_message_message(int server, struct smessage *mess);

/* A global to say if we are running or not. */

static char *scope_names[NUM_SCOPES] = {"S1", "S2", "E1", "E2", "W1", "W2"};

/************************************************************************/
/* start_chara_server()							*/
/*									*/
/* DO all the things we need to do to be a chara server.		*/
/* Returns 0 if all goes well, non-zero otherwise.			*/
/************************************************************************/

int start_chara_server(void)
{
	/* Initialise the jobs */

	server_init_jobs();

	/* 
 	 * Open the server socket. This routine will select a socket
 	 * and register the server with the socket manager. The first argument
 	 * is the name of the server. The second is the command line 
 	 * argument to type to restart the server. This is often most useful
 	 * for CHARA specific servers and for most beam combiners will be
 	 * left empty. This function also sets up message functions for 
 	 * all standard messages such as NOOP etc.
 	 */

	if (auto_open_message_socket("champ","") != NOERROR)
	{
		error(ERROR,"Failed to connect to socket manager.");
		return -1;
	}
	
	/*
	 * Add any local message jobs. Each one associates a function
	 * with a message. You will need one of these for all messages that
	 * your server will need to deal with.
	 */

        server_add_message_job(ASTROMOD_SET_USE_ASTROMOD_LIB, 
			message_astromod_set_use_astromod_lib);
        server_add_message_job(OBSERVE_OPLE_NUMBER, 
			message_observe_ople_number);
        server_add_message_job(ASTROMOD_INIT, message_astromod_init);
        server_add_message_job(ASTROMOD_SET_REFERENCE_CART, 
			message_astromod_set_reference_cart);
        server_add_message_job(ASTROMOD_WHICH_REFERENCE_CART, 
			message_astromod_which_reference_cart);
        server_add_message_job(ASTROMOD_SET_ACTIVE_SCOPES, 
			message_astromod_set_active_scopes);
        server_add_message_job(ASTROMOD_SET_STAR, message_astromod_set_star);
        server_add_message_job(OBSERVE_STAR_ACQUIRED, 
			message_observe_star_acquired);
        server_add_message_job(ASTROMOD_SET_OBS_WAVELENGTH, 
			message_astromod_set_obs_wavelength);

	/* Set the name */

	set_server_name("CHAMP");

	/*
 	 * This puts things into "standalone" mode, which means it wil
	 * not check the computer name of connecting machines. If set to
	 * TRUE, it will only allow machines chara-array.org to connect,
	 * otherwise any machine can connect.
	 * Note that "localhost" does not pass the chara-array.org test.
	 */

	set_standalone(TRUE);

	/* That is all */

	return 0;

} /* start_chara_server() */

/************************************************************************/
/* stop_chara_server()							*/
/*									*/
/* Stop our services to the CHARA system.				*/
/************************************************************************/

void stop_chara_server(void)
{

	/*
	 * OK, we should close things down gracefully. 
	 * that will close the socket, deregister with the socketmanager
	 * send a message to all clients that we are closing and so on.
	 */

	close_server_logfile();
	send_server_stopping();
	close_message_socket();
        error(MESSAGE,"Closing down as CHAMP server.");

} /* stop_chara_server() */

/************************************************************************/
/* message_astromod_set_use_astromod_lib()                              */
/*                                                                      */
/* Example message function.						*/
/* Returns error level.                                                 */
/************************************************************************/

int message_astromod_set_use_astromod_lib(struct smessage *message)
{
        if (message->length != sizeof(int))
        {
                return error(ERROR,
		"Received ASTROMOD_SET_USE_ASTROMOD_LIB with bad data.");
        }

	if (*((int *) message->data))
		error(MESSAGE,"Using astrmod library.");
	else
		error(MESSAGE,"Not using astrmod library.");

	return NOERROR;

} /* message_astromod_set_use_astromod_lib() */

/************************************************************************/
/* message_observe_ople_number()	                                */
/*                                                                      */
/* Example message function.						*/
/* Returns error level.                                                 */
/************************************************************************/

int message_observe_ople_number(struct smessage *message)
{
        if (message->length != sizeof(int))
        {
                return error(ERROR,
		"Received OBSERVE_OPLE_NUMBER with bad data.");
        }

	error (MESSAGE, "OPLE number is %d", *((int *) message->data));

	return NOERROR;

} /* message_observe_ople_number() */

/************************************************************************/
/* message_astromod_init() 		                                */
/*                                                                      */
/* Example message function.						*/
/* Returns error level.                                                 */
/************************************************************************/

int message_astromod_init(struct smessage *message)
{
	struct s_astromod_init *data;
	int i;

        if (message->length != sizeof(*data))
        {
                return error(ERROR,
			"Received ASTROMOD_INIT with bad data.");
        }

	data = (struct s_astromod_init *)message->data;

	error(MESSAGE,"New configuration sent.");
	for(i=0; i<NUM_SCOPES; i++)
		error(MESSAGE,"%s BEAM %d POP %d",
			scope_names[i], data->beams[i]+1, data->pops[i]+1);

	return NOERROR;

} /* message_astromod_init() */

/************************************************************************/
/* message_astromod_set_reference_cart()                                 */
/*                                                                      */
/* Example message function.						*/
/* Returns error level.                                                 */
/************************************************************************/

int message_astromod_set_reference_cart(struct smessage *message)
{
	int	new_reference_cart;

        if (message->length != sizeof(new_reference_cart))
        {
                return error(ERROR,
			"Received ASTROMOD_SET_REFERENCE CART with bad data.");
        }

	/* Here we just copy the data into a local global */
	
	new_reference_cart = *((int *)message->data);

	/* 
	 * And we tell the world about this. Note we use the "error" function
	 * which works exactly like printf and has an extra argument which
	 * can be ERROR, WARNING, MESSAGE or NOERROR.
	 */

	if (new_reference_cart < 0 || new_reference_cart > 5)
		return error(ERROR, "New reference %d out of range.",
			new_reference_cart);
	else
		error(MESSAGE,"Reference cart set to %s by %s.",
				scope_names[new_reference_cart],
				active_machine);

	reference_cart = new_reference_cart;

	/* All is well so we return NOERROR */

        return NOERROR;

} /*  message_astromod_set_reference_cart() */

/************************************************************************/
/* message_astromod_which_reference_cart()                              */
/*                                                                      */
/* Example message function.						*/
/* Returns error level.                                                 */
/************************************************************************/

int message_astromod_which_reference_cart(struct smessage *message)
{
	struct smessage mess;

        if (message->length != 0)
        {
                return error(ERROR,
		"Received ASTROMOD_WHICH_REFERENCE CART with bad data.");
        }

        mess.type = ASTROMOD_SET_REFERENCE_CART;
        mess.length = sizeof(reference_cart);
        mess.data = (unsigned char *)&reference_cart;

	/*
	 * We send a reply to the "active socket", that is the socket
 	 * that sent the message we are responding to.
 	 */

        if (server_send_message(active_socket, &mess) != NOERROR)
        {
            error(ERROR,"Failed to send ASTROMOD_SET_REFERENCE_CART.");
        }

        return NOERROR;

} /*  message_astromod_which_reference_cart() */

/************************************************************************/
/* message_astromod_set_active_scopes()	                                */
/*                                                                      */
/* Example message function.						*/
/* Returns error level.                                                 */
/************************************************************************/

int message_astromod_set_active_scopes(struct smessage *message)
{
	int	i;
	char	*data;

        if (message->length != NUM_SCOPES * sizeof(*data))
        {
                return error(ERROR,
		"Received ASTROMOD_SET_ACTIVE_SCOPES with bad data.");
        }

	data = (char *)message->data;

	for(i=0; i<NUM_SCOPES; i++)
	{
	    if (data[i])
		error(MESSAGE,"Scope %s is active.", scope_names[i]);
	    else
		error(MESSAGE,"Scope %s is not active.", scope_names[i]);
	}

	return NOERROR;

} /* message_astromod_set_active_scopes() */

/************************************************************************/
/* message_astromod_set_star()		                                */
/*                                                                      */
/* Example message function.						*/
/* Returns error level.                                                 */
/************************************************************************/

int message_astromod_set_star(struct smessage *message)
{
        if (message->length != sizeof(int))
        {
                return error(ERROR,
		"Received ASTROMOD_SET_STAR with bad data.");
        }

	error (MESSAGE, "CHARA number for star is %d",*((int *) message->data));

	if(star.chara_number > 0)
	  {
	    if (!get_star_by_number(*((int *) message->data), &star))
	      {
		error(ERROR, "No star found in the database");
		return ERROR;
	      }
	    else
	      {
		error(MESSAGE, "Got star %s from database", star.name);
	      }
	  }
	else
	  error(MESSAGE, "Star number < 0", star.name);
	return NOERROR;

} /* message_astromod_set_star() */

/************************************************************************/
/* message_observe_star_acquired()	                                */
/*                                                                      */
/* Example message function.						*/
/* Returns error level.                                                 */
/************************************************************************/

int message_observe_star_acquired(struct smessage *message)
{
        if (message->length != 0)
        {
                return error(ERROR,
		"Received OBSERVE_STAR_ACQUIRED with bad data.");
        }

	error (MESSAGE, "The star has been acquired.");

	return NOERROR;

} /* message_observe_star_acquired() */

/************************************************************************/
/* message_astromod_set_obs_wavelength()	                        */
/*                                                                      */
/* Example message function.						*/
/* Returns error level.                                                 */
/************************************************************************/

int message_astromod_set_obs_wavelength(struct smessage *message)
{
        if (message->length != sizeof(double))
        {
                return error(ERROR,
		"Received OBSERVE_SET_OBS_WAVELENGTH with bad data.");
        }

	error (MESSAGE, "Observing wavelength is %lf microns.",
		*((double *)message->data));

	return NOERROR;

} /* message_astromod_set_obs_wavelength() */

/************************************************************************/
/* client message_message()                                             */
/*                                                                      */
/* A message to display a message.                                      */
/* This is for when we are behaving like a client.			*/
/************************************************************************/

int client_message_message(int server, struct smessage *mess)
{
        char *s,*p1,*p2;
        int     i;

        if (mess->length == 0)
        {
                return TRUE;
        }
        else
        {
                if ((s = malloc(mess->length+1)) == NULL)
                {
                        error(ERROR,
                                "Ran out of memory getting message string.");
			return ERROR;
                }

                for(i=0, p1=s, p2=(char *)mess->data;
                    i < mess->length; i++, p1++, p2++) *p1 = *p2;
                *p1 = 0;

                error(MESSAGE, "Message from server: %s", s);

                free(s);
        }

        return TRUE;

} /* client_message_message() */

